/***************************************************************************
 * LPRng - An Extended Print Spooler System
 *
 *
 * ipp events and subscriptions
 * copyright 2021 Vaclav Michalek, Joint Laboratory of Optics, Olomouc
 *
 * See LICENSE for conditions of use.
 *
 ***************************************************************************/


#include "lp.h"
#include "linelist.h"
#include "event.h"
#include "ipp.h"
#include "ipp_events.h"
#include "getqueue.h"

/* 
#include "linelist.h"
#include "utilities.h" */





int Ips_remove_subscription(const char *printer, const char *ppath, int sub_id)
{

	struct ipp_subs *p, *r;
	struct ipp_event *e, *f;
	struct int_list *l;
	int s, rs, st;
	char buf[SMALLBUFFER];


	rs = CLIENT_ERROR_NOT_FOUND;
	p = ipp_subs_list;


	DEBUG2("Ips_remove_subscription: subscription id %d printer %s ppath %s", sub_id, printer ? printer : "<NULL>", ppath ? ppath : "<NULL>" );
	/*for (p = ipp_subs_list; p; p = p->next) {
		DEBUG2("Ips_remove_subscription: start list sub id %d printer %s ppath %s", p->id, p->printer ? p->printer : "<NULL>", p->ppath ? p->ppath : "<NULL>" );
	}*/

	p = ipp_subs_list; r = NULL;
	while(p) {
		s = 0;
		if ((p->expire < time(0)) ||
		    (s = (sub_id && sub_id == p->id /*&& !safestrcmp(p->printer, printer) && !safestrcmp(p->ppath, ppath)*/))
		   )
		{
			if (s) {
				DEBUG2("Ips_remove_subscription: subscription id %d cancel requested, removing", p->id);
				rs = SUCCESSFUL_OK;
			} else 
				DEBUG2("Ips_remove_subscription: subscription id %d expired, removing", p->id);
			if (r) r->next = p->next; /*no change r*/
			else ipp_subs_list = p->next;
			if (p->printer) free(p->printer);
			if (p->ppath) free(p->ppath);
			if (p->username) free(p->username);
			if (p->user_data) free(p->user_data);
			/*remove all events*/
			while ((e = p->events)) {
				p->events = p->events->next;
				if (e->printer) free(e->printer);
				if (e->reason) free(e->reason);
				if (e->xname) free(e->xname);
				free(e);
			}
			while ((l = p->clients)) {
				p->clients = p->clients->next;
				/*stop running notifications*/
				buf[0] = '\0';
				EddIntData(buf, sizeof(buf), E_SUBS_CNTF);
				EddIntData(buf, sizeof(buf), time(0));
				EddIntData(buf, sizeof(buf), SUCCESSFUL_OK_EVENTS_COMPLETED);
				EddIntData(buf, sizeof(buf), p->id); /*removed subscription id*/
				EddLf(buf, sizeof(buf));
				st = write(l->i, buf, safestrlen(buf));
				DEBUG2("Ips_remove_subscription: send waitmode end fd %d write bytes %d errno %d", l->i, st >=0 ? st : 0, st >= 0 ? 0 : errno);
				free(l);
			}
			free(p);
			if (r) p = r->next; else p = ipp_subs_list;
		} else {
			/*remove only events expired*/
			e = p->events; f = 0;
			while (e) {
				if (e->printer_up_time + NTF_EVENT_LIFE < time(0)) {
					DEBUG3("Ips_remove_subscription: remove expired event id %d e seqno %d", p->id, e->seqno);
					if (f) {
						f->next = e->next;
						if (e->printer) free (e->printer);
						if (e->reason) free (e->reason);
						if (e->xname) free (e->xname);
						free(e);
						e = f->next;
					} else {
						p->events = e->next;
						if (e->printer) free (e->printer);
						if (e->reason) free (e->reason);
						if (e->xname) free (e->xname);
						free(e);
						e = p->events;
					}
				} else {
					f = e;
					e = e->next;
				}
			}
			r = p; p = p->next;
		}
	}

	/*for (p = ipp_subs_list; p; p = p->next) {
		DEBUG2("Ips_remove_subscription: end list sub id %d printer %s ppath %s", p->id, p->printer ? p->printer : "<NULL>", p->ppath ? p->ppath : "<NULL>" );
	}
	if (!ipp_subs_list) DEBUG2("Ips_remove_subscription: list sub empty" );*/

	return rs;
}

int Ips_delete_wait_client(int fd)
{
	struct ipp_subs *sub;
	struct int_list *clnt, *prev;
	int rs = 0;

	for (sub = ipp_subs_list; sub; sub = sub->next) {
		clnt = sub->clients; prev = NULL;
		while (clnt) {
			if (clnt->i == fd) {
				if (prev) {
					prev->next = clnt->next;
					free(clnt);
					clnt = prev->next;
				} else {
					sub->clients = clnt->next;
					free(clnt);
					clnt = sub->clients;
				}
			} else {
				prev = clnt;
				clnt = clnt->next;
			}
			++rs;
			DEBUG2("Ips_delete_wait_client: sub %d fd %d unregistered", sub->id, fd);
		}
	}

	return rs;
}

int Ips_subscription_create(struct line_list *params)
{

	int id, lease, status, fd, trans;
	struct ipp_subs *sub, *p;
	char resp[SMALLBUFFER];

	id = 0;
	lease = 0;
	status = SUCCESSFUL_OK;
	fd = -1;
	trans = 0;

	/*struct ipp_subs *r;
	for (p = ipp_subs_list, r = NULL; p; r = p, p = p->next) {
		DEBUG2("Ips_subscription_create: start list sub id %d printer %s ppath %s", p->id, p->printer ? p->printer : "<NULL>", p->ppath ? p->ppath : "<NULL>" );
	}*/

	if(DEBUGL2)Dump_line_list("Ips_subscription_create: params", params);

	if (params->count < 3) {
		DEBUG2("Ips_subscription_create: too small number of params (<3)");
		return -1;
	}
	fd = strtol(params->list[1], NULL, 0);
	if (params->count < 9) {
		DEBUG2("Ips_subscription_create: too small number of params (<9)");
		status = SERVER_ERROR_INTERNAL_ERROR;
		goto endproc;
	}
	trans = strtol(params->list[2], NULL, 0);

	/*count and find id*/
	sub = ipp_subs_list; p = NULL; id = 1;
	while (sub) {
		if (sub->id > id) break;
		p = sub;
		sub = sub->next;
		++id;
	}
	if (id > NTF_MAX_SUBSCRIPTIONS) {
		status = CLIENT_ERROR_TOO_MANY_SUBSCRIPTIONS;
		goto endproc;
	}

	sub = malloc_or_die(sizeof(struct ipp_subs), __FILE__, __LINE__);
	sub->next = p ? p->next : ipp_subs_list;
	if (p) p->next = sub; else ipp_subs_list = sub;

	sub->id = id;
	sub->clients = NULL;
	ExmDataHex(&(sub->printer), params->list[3]);
	ExmDataHex(&(sub->ppath), params->list[4]);
	ExmDataHex(&(sub->username), params->list[5]);
	sub->seqno = 0;
	lease = strtol(params->list[6], NULL, 0);
	if (lease <= NTF_LEASE_DURATION_MIN) lease = NTF_LEASE_DURATION_MIN;
	if (lease > NTF_LEASE_DURATION_MAX) lease = NTF_LEASE_DURATION_MAX;
	sub->expire = time(0) + lease;
	sub->lease = lease;
	sub->event_types = strtol(params->list[7], NULL, 0);
	sub->ulen =  ExmDataHex(&(sub->user_data), params->list[8]);
	sub->events = NULL;

	DEBUG2("Ips_subscription_create: printer %s path %s username %s subscription id %d events 0x%x lease %d",
	         sub->printer, sub->ppath, sub->username, sub->id, sub->event_types, lease );

    endproc:

	memset(resp, 0, sizeof(resp));
	EddIntData(resp, sizeof(resp), E_SUBS_CREATE);
	EddIntData(resp, sizeof(resp), trans);
	EddIntData(resp, sizeof(resp), status);
	EddIntData(resp, sizeof(resp), id);
	EddIntData(resp, sizeof(resp), lease);
	EddLf(resp, sizeof(resp));
	DEBUG1("Ips_subscription_create: subscription operation status %d", status);

	/*send response back*/
	status = write(fd, resp, safestrlen(resp));
	DEBUG2("Ips_subscription_create: send response fd %d write bytes %d errno %d", fd, status >=0 ? status : 0, status >= 0 ? 0 : errno);

	/*for (p = ipp_subs_list, r = NULL; p; r = p, p = p->next) {
		DEBUG2("Ips_subscription_create: end list sub id %d printer %s ppath %s", p->id, p->printer ? p->printer : "<NULL>", p->ppath ? p->ppath : "<NULL>" );
	}*/

	return 0;

}


int Ips_subscription_attributes(struct line_list *params)
{

	int sub_id, job_id, limit, mysubs, status, fd, trans;
	char *printer, *ppath;
	struct ipp_subs *sub;
	char *user, *resp;
	int count, sz;

	status = SUCCESSFUL_OK;
	fd = -1;
	trans = 0;
	resp = NULL;
	user = NULL;
	count = 0;
	printer = NULL;
	ppath = NULL;


	if(DEBUGL2)Dump_line_list("Ips_subscription_attributes: params", params);

	if (params->count < 3) {
		DEBUG2("Ips_subscription_attributes: too small number of params (<3)");
		return -1;
	}
	fd = strtol(params->list[1], NULL, 0);
	sz = 2*sizeof(size_t) + 1 + 3*(3*sizeof(int) + 1) + 2; /*size, E_SUBS_ATTRS, trans, status, lf\0*/

	if (params->count < 10) {
		DEBUG2("Ips_subscription_attributes: too small number of params (<10)");
		status = SERVER_ERROR_INTERNAL_ERROR;
		resp = malloc_or_die(sz, __FILE__, __LINE__); resp[0] = '\0';
		EddIntData(resp, sz, E_SUBS_ATTRS);
		EddIntData(resp, sz, trans);
		EddIntData(resp, sz, status);
		goto endproc;
	}
	trans = strtol(params->list[2], NULL, 0);
	ExmDataHex(&printer, params->list[3]);
	ExmDataHex(&ppath, params->list[4]);
	ExmDataHex(&user, params->list[5]);
	sub_id = strtol(params->list[6], NULL, 0);
	job_id = strtol(params->list[7], NULL, 0);
	limit = strtol(params->list[8], NULL, 0);
	mysubs = strtol(params->list[9], NULL, 0);

	if (job_id) {
		DEBUG2("Ips_subscription_attributes: jobid defined, none subscription");
		/*status ok*/
		resp = malloc_or_die(sz, __FILE__, __LINE__); resp[0] = '\0';
		EddIntData(resp, sz, E_SUBS_ATTRS);
		EddIntData(resp, sz, trans);
		EddIntData(resp, sz, status);
		goto endproc;
	}

	/*compute storage size*/
	sub = ipp_subs_list; count = 0;
	while (sub) {
		if ((sub_id && sub->id != sub_id) ||
		    safestrcmp(sub->printer, printer) ||
		    safestrcmp(sub->ppath, ppath))  continue;
		if (mysubs && safestrcmp(user, sub->username)) continue;
		count++;

		sz += 3*sizeof(int) + 1; /*subid, */
		sz += 3*sizeof(int) + 1; /*seqno */
		sz += 3*sizeof(time_t) + 1; /*expiration */
		sz += 2*safestrlen(sub->printer) + 1;
		sz += 2*safestrlen(sub->username) + 1;
		sz += 3*sizeof(int) + 1; /*events */
		sz += 2*sub->ulen + 2; /*user data */
		sz += 3*sizeof(int) + 1; /*lease duration */

		if (sub_id) break;
		if (limit && limit >= count) break;
		sub = sub->next;
	}

	resp = malloc_or_die(sz, __FILE__, __LINE__);
	memset(resp, 0, sz); /*resp[0] = '\0';*/
	/*fill response*/
	EddIntData(resp, sz, E_SUBS_ATTRS);
	EddIntData(resp, sz, trans);
	EddIntData(resp, sz, status);
	sub = ipp_subs_list; count = 0;
	while (sub) {
		if ((sub_id && sub->id != sub_id) ||
		    safestrcmp(sub->printer, printer) ||
		    safestrcmp(sub->ppath, ppath))  continue;
		if (mysubs && safestrcmp(user, sub->username)) continue;
		count++;

		EddIntData(resp, sz, sub->id);
		EddIntData(resp, sz, sub->seqno);
		EddHexData(resp, sz, (const char *)&(sub->expire), sizeof(time_t));
		EddHexData(resp, sz, sub->printer, safestrlen(sub->printer));
		EddHexData(resp, sz, sub->ppath, safestrlen(sub->ppath));
		EddHexData(resp, sz, sub->username, safestrlen(sub->username));
		EddIntData(resp, sz, sub->event_types);
		EddHexData(resp, sz, sub->user_data, sub->ulen);
		EddIntData(resp, sz, sub->lease);

		if (sub_id) break;
		if (limit && limit >= count) break;
		sub = sub->next;
	}

    endproc:

	if (user) free(user);
	if (printer) free(printer);
	if (ppath) free(ppath);

	EddLf(resp, sz);
	DEBUG2("Ips_subscription_attributes: response: %s", resp);
	DEBUG1("Ips_subscription_attributes: operation status %d, subscription count %d", status, count);
	/*send response back*/
	status = write(fd, resp, safestrlen(resp));
	DEBUG2("Ips_subscription_attributes: send response fd %d write bytes %d errno %d", fd, status >=0 ? status : 0, status >= 0 ? 0 : errno);

	if (resp) free(resp);

	return 0;
}


int Ips_subscription_renew(struct line_list *params)
{

	int sub_id, lease, status, fd, trans;
	char *printer, *ppath;
	struct ipp_subs *sub;
	char resp[SMALLBUFFER];

	status = SUCCESSFUL_OK;
	fd = -1;
	trans = 0;
	lease = 0;
	printer = NULL;
	ppath = 0;


	if(DEBUGL2)Dump_line_list("Ips_subscription_renew: params", params);

	if (params->count < 3) {
		DEBUG2("Ips_subscription_renew: too small number of params (<3)");
		return -1;
	}
	fd = strtol(params->list[1], NULL, 0);

	if (params->count < 8) {
		DEBUG2("Ips_subscription_renew: too small number of params (<8)");
		status = SERVER_ERROR_INTERNAL_ERROR;
		goto endproc;
	}
	trans = strtol(params->list[2], NULL, 0);
	ExmDataHex(&printer, params->list[3]);
	ExmDataHex(&ppath, params->list[4]);
	sub_id = strtol(params->list[6], NULL, 0);
	lease = strtol(params->list[7], NULL, 0);

	status = CLIENT_ERROR_NOT_FOUND;

	/*find and modify subscription*/
	for (sub = ipp_subs_list; sub; sub = sub->next) {
		if ((sub->id == sub_id) &&
		   !safestrcmp(sub->printer, printer) &&
		   !safestrcmp(sub->printer, ppath) )
		{
			if ((lease <= NTF_LEASE_DURATION_MIN) || (lease > NTF_LEASE_DURATION_MAX))  {
				lease = NTF_LEASE_DURATION;
				status = SUCCESSFUL_OK_IGNORED_SUBSTITUED;
			} else status = SUCCESSFUL_OK;
			sub->expire = time(0) + lease;
			sub->lease = lease;
			DEBUG2("Ips_subscription_attributes: renewed subscription id %d lease %d", sub->id, lease);
			break;
		}
	}

    endproc:

	memset(resp, 0, sizeof(resp));
	EddIntData(resp, sizeof(resp), E_SUBS_RENEW);
	EddIntData(resp, sizeof(resp), trans);
	EddIntData(resp, sizeof(resp), status);
	EddIntData(resp, sizeof(resp), lease);
	EddLf(resp, sizeof(resp));

	DEBUG2("Ips_subscription_renew: status %d response %s", status, resp);
	/*send response back*/
	status = write(fd, resp, safestrlen(resp));
	DEBUG2("Ips_subscription_renew: send response fd %d write bytes %d errno %d", fd, status >=0 ? status : 0, status >= 0 ? 0 : errno);

	if(printer) free(printer);
	if(ppath) free(ppath);

	return 0;
}


int Ips_subscription_cancel(struct line_list *params)
{
	int fd, trans, sub_id, status;
	char *printer, *ppath, resp[SMALLBUFFER];


	if(DEBUGL2)Dump_line_list("Ips_subscription_cancel: params", params);

	if (params->count < 3) {
		DEBUG2("Ips_subscription_cancel: too small number of params (<3)");
		return -1;
	}
	fd = strtol(params->list[1], NULL, 0);

	if (params->count < 7) {
		DEBUG2("Ips_subscription_renew: too small number of params (<7)");
		status = SERVER_ERROR_INTERNAL_ERROR;
		goto endproc;
	}
	trans = strtol(params->list[2], NULL, 0);
	ExmDataHex(&printer, params->list[3]);
	ExmDataHex(&ppath, params->list[4]);
	sub_id = strtol(params->list[6], NULL, 0);

	status = Ips_remove_subscription(printer, ppath, sub_id);

    endproc:

	memset(resp, 0, sizeof(resp));
	EddIntData(resp, sizeof(resp), E_SUBS_CANCEL);
	EddIntData(resp, sizeof(resp), trans);
	EddIntData(resp, sizeof(resp), status);
	EddLf(resp, sizeof(resp));

	DEBUG2("Ips_subscription_cancel: status %d response %s", status, resp);
	/*send response back*/
	status = write(fd, resp, safestrlen(resp));
	DEBUG2("Ips_subscription_cancel: send response fd %d write bytes %d errno %d", fd, status >=0 ? status : 0, status >= 0 ? 0 : errno);

	if(printer) free(printer);
	if(ppath) free(ppath);

	return 0;
}


int Ips_event(struct line_list *params) {


	char *printer, *reason, *xname;
	char *resp;
	int event, state, id, count, rsz, st;
	struct ipp_subs *sub;
	struct ipp_event *ev;
	struct int_list *cl;

	if (params->count < 6) {
		DEBUG2("Ips_event: too small number of params (<6)");
		return -1;
	}

	event = strtol(params->list[0], NULL, 0);
	ExmDataHex(&printer, params->list[1]);
	state = strtol(params->list[2], NULL, 0);
	ExmDataHex(&reason, params->list[3]);
	id = strtol(params->list[4], NULL, 0);
	ExmDataHex(&xname, params->list[5]);

	DEBUG1("Ips_event: IPP %d (%s) event occured, printer %s state %d reason %s aux_id %d ",
	        event, event == E_PRS_CHANGE ? "printer": "job", printer, state, reason, id);

	if (!ipp_subs_list) goto endproc;

	for (count = 0, sub = ipp_subs_list; sub; sub = sub->next, ++count) {
		if (count >= NTF_MAX_EVNT_COUNT) {
			DEBUG2("Ips_event: Event NOT added to subscription %d, too many events", sub->id);
			break;
		}
		if (safestrcmp(sub->printer, printer) && safestrcmp(sub->printer, IPPC_ALLPRINTERS) ) continue;
		if ( ((event == E_PRS_CHANGE) && (sub->event_types & IEVENT_PR_STATECHANGE)) ||
		     ((event == E_PRS_CHANGE) && (sub->event_types & IEVENT_PR_STOPPED) && state == IPPAV_PRS_STOPPED) ||
		     ((event == E_JOB_CREATE) && (sub->event_types & (IEVENT_JOB_STATECHANGE | IEVENT_JOB_CREATED))) ||
		     ((event == E_JOB_CHANGE) && (sub->event_types & IEVENT_JOB_STATECHANGE)) ||
		     ((event == E_JOB_CHANGE) && (sub->event_types & IEVENT_JOB_COMPLETED) &&
		                                 (state == IPPAV_JRS_COMPLETED || state == IPPAV_JRS_ABORTED || state == IPPAV_JRS_CANCELED ))
		   ) {
			/*add printer/job event*/
			if (!sub->events) { 
				sub->events = malloc_or_die(sizeof(struct ipp_event), __FILE__, __LINE__);
				ev = sub->events;
			} else {
				for (ev = sub->events; ev->next; ev = ev->next);
				ev->next = malloc_or_die(sizeof(struct ipp_event), __FILE__, __LINE__);
				ev = ev->next;
			}
			ev->seqno = ++sub->seqno;
			if (event == E_PRS_CHANGE) ev->etype = (sub->event_types & IEVENT_PR_STATECHANGE) ? IEVENT_PR_STATECHANGE : IEVENT_PR_STOPPED;
			else if (event == E_JOB_CREATE) ev->etype = (sub->event_types & IEVENT_JOB_CREATED) ? IEVENT_JOB_CREATED : IEVENT_JOB_STATECHANGE;
			else ev->etype = ( (sub->event_types & IEVENT_JOB_COMPLETED) && (state == IPPAV_JRS_COMPLETED || state == IPPAV_JRS_ABORTED || state == IPPAV_JRS_CANCELED) ) ?
			              IEVENT_JOB_COMPLETED : IEVENT_JOB_STATECHANGE;
			ev->printer = safestrdup(printer, __FILE__, __LINE__);
			ev->printer_up_time = time(0);
			ev->state = state;
			ev->aux_id = id;
			ev->reason = safestrdup(reason, __FILE__, __LINE__);
			ev->xname = safestrdup(xname, __FILE__, __LINE__);
			ev->next = 0;
			/*if running notifications, send event to the clients*/
			resp = NULL;
			for (cl = sub->clients; cl; cl = cl->next) {
				if (!resp) {
					/*see also Ips_get_notifications*/
					rsz = 2*sizeof(size_t) + 1 + 4*(3*sizeof(int) + 1) + 2; /*size, E_SUBS_GNTF, trans, status, waitmode? lf\0*/
					rsz += 3*sizeof(int) + 1 /*subscription id*/
					     + 2*safestrlen(ev->printer) + 2
					     + 2*safestrlen(sub->ppath) + 2
					     + 3*sizeof(int) +1 /*event*/
					     + 2*sizeof(time_t) + 1 /*event time*/
					     + 3*sizeof(int) + 1 /*seqno*/
					     + 2*(sub->ulen) + 2 /*user data*/
					     + 3*sizeof(int) + 1 /*aux-id*/
					     + 3*sizeof(int) + 1 /*state*/
					     + 2*safestrlen(ev->reason) + 2 /*reason*/
					     + 2*safestrlen(ev->xname) + 2; /*job name*/
					resp = malloc_or_die(rsz, __FILE__, __LINE__); resp[0] = '\0';
					EddIntData(resp, rsz, E_SUBS_GNTF);
					st = time(0);         EddIntData(resp, rsz, st);
					st = SUCCESSFUL_OK;   EddIntData(resp, rsz, st);
					st = 0;               EddIntData(resp, rsz, st);
					EddIntData(resp, rsz, sub->id); /*subscription id*/
					EddHexData(resp, rsz, ev->printer, safestrlen(ev->printer));
					EddHexData(resp, rsz, sub->ppath, safestrlen(sub->ppath));
					EddIntData(resp, rsz, ev->etype);
					EddHexData(resp, rsz, (const char *)&(ev->printer_up_time), sizeof(time_t));
					EddIntData(resp, rsz, sub->seqno); /*rfc3996 says seqno from subscription, not event  EddIntData(resp, sz, ev->seqno);*/
					EddHexData(resp, rsz, sub->user_data, sub->ulen);
					EddIntData(resp, rsz, ev->aux_id);
					EddIntData(resp, rsz, ev->state);
					EddHexData(resp, rsz, ev->reason, safestrlen(ev->reason));
					EddHexData(resp, rsz, ev->xname, safestrlen(ev->xname));
					EddLf(resp, rsz);
				}
				st = write(cl->i, resp, safestrlen(resp));
				DEBUG2("Ips_event: send event notification to fd %d write bytes %d errno %d", cl->i, st>=0 ? st: 0, st >= 0 ? 0 : errno);
			}
			if (resp) free(resp);
		}
	}


    endproc:

	if (printer) free(printer);
	if (reason) free(reason);
	if (xname) free(xname);

	/*Debug list subs and events*/
	for (sub = ipp_subs_list; sub; sub = sub->next) {
		DEBUG2("Ips_event: list sub id %d printer %s ppath %s", sub->id, sub->printer ? sub->printer : "<NULL>", sub->ppath ? sub->ppath : "<NULL>" );
		for (ev = sub->events; ev; ev = ev->next) {
			/*DEBUG2("Ips_event: list sub id %d event seqno %d printer %s src %s state %d id %d xname %s",
			        sub->id, ev->seqno, ev->printer ? ev->printer : "<NULL>", ev->etype & (IEVENT_PR_STATECHANGE | IEVENT_PR_STOPPED) ? "printer" : "job",
			         ev->state, ev->aux_id, ev->xname ? ev->xname : "<NULL>" );*/
		}
		if (!sub->events) DEBUG2("Ips_event: list sub id %d no events", sub->id);
	}
	if (!ipp_subs_list) DEBUG2("Ips_event: list sub empty" );


	return 0;
}

int Ips_get_notifications(struct line_list *params) {

	int i, fd, trans, waitinterval, sz, id, seqn, status, evcount;
	/*char *printer, *ppath;*/
	char *resp;
	struct ipp_subs *sub;
	struct ipp_event *ev;
	struct int_list *cl;


	if(DEBUGL3)Dump_line_list("Ips_get_notifications: params", params);

	if (params->count < 3) {
		DEBUG2("Ips_get_notifications: too small number of params (<3)");
		return -1;
	}
	resp = NULL;
	fd = strtol(params->list[1], NULL, 0);
	sz = 2*sizeof(size_t) + 1 + 4*(3*sizeof(int) + 1) + 2; /*size, E_SUBS_GNTF, trans, status, waitmode? lf\0*/
	evcount = 0;

	if (params->count < 8) {
		DEBUG2("Ips_get_notifications: too small number of params (<8)");
		status = SERVER_ERROR_INTERNAL_ERROR;
		trans = 0;
		resp = malloc_or_die(sz, __FILE__, __LINE__); resp[0] = '\0';
		EddIntData(resp, sz, E_SUBS_GNTF);
		EddIntData(resp, sz, trans);
		EddIntData(resp, sz, status);
		EddIntData(resp, sz, NTF_MIN_WAIT);
		goto endproc;
	}
	trans = strtol(params->list[2], NULL, 0);
	/*unused
	ExmDataHex(&printer, params->list[3]);
	ExmDataHex(&ppath, params->list[4]);
	*/
	waitinterval = strtol(params->list[6], NULL, 0);
	if (waitinterval) waitinterval = Lpq_status_interval_DYN > NTF_MIN_WAIT ? Lpq_status_interval_DYN : NTF_MIN_WAIT;

	status = CLIENT_ERROR_NOT_FOUND;

	/*count size for response*/
	for (i = 7; i < params->count; ++i) {
		id = strtol(params->list[i], NULL, 0);
		seqn = strtol(params->list[++i], NULL, 0);

		DEBUG4("Ips_get_notifications: id %d seqno %d", id, seqn);

		for (sub = ipp_subs_list; sub; sub = sub->next) {
			if (sub->id == id) {
				for(ev = sub->events; ev; ev = ev->next) {
				DEBUG2("Ips_get_notifications: sub id %d, seqno %d rseq %d event seqno %d", sub->id, sub->seqno, seqn, ev->seqno);
					if (ev->seqno >= seqn) {
						sz += 3*sizeof(int) + 1; /*subscription id*/
						sz += 2*safestrlen(ev->printer) + 2;
						sz += 2*safestrlen(sub->ppath) + 2;
						sz += 3*sizeof(int) +1; /*event*/
						sz += 2*sizeof(time_t) + 1; /*event time*/
						sz += 3*sizeof(int) + 1; /*seqno*/
						sz += 2*(sub->ulen) + 2; /*user data*/
						sz += 3*sizeof(int) + 1; /*aux-id*/
						sz += 3*sizeof(int) + 1; /*state*/
						sz += 2*safestrlen(ev->reason) + 2; /*reason*/
						sz += 2*safestrlen(ev->xname) + 2; /*job name*/
						/*status ? not needed yet*/
						++evcount;
					}
				}
				if (status == CLIENT_ERROR_NOT_FOUND) status = SUCCESSFUL_OK;
				goto subc_found;
			}
		}
		if (status == SUCCESSFUL_OK) status = SUCCESSFUL_OK_IGNORED_SUBSTITUED;
	  subc_found:
		;
	}

	resp = malloc_or_die(sz, __FILE__, __LINE__);
	resp[0] = '\0';
	/*fill response*/
	EddIntData(resp, sz, E_SUBS_GNTF);
	EddIntData(resp, sz, trans);
	EddIntData(resp, sz, status);
	EddIntData(resp, sz, waitinterval);
	/*count size for response*/
	for (i = 7; i < params->count; ++i) {
		id = strtol(params->list[i], NULL, 0);
		seqn = strtol(params->list[++i], NULL, 0);
		for (sub = ipp_subs_list; sub; sub = sub->next) {
			if (sub->id == id) {
				for(ev = sub->events; ev; ev = ev->next) {
					if (ev->seqno >= seqn) {
						/*see also Ips_event*/
						EddIntData(resp, sz, sub->id); /*subscription id*/
						EddHexData(resp, sz, ev->printer, safestrlen(ev->printer));
						EddHexData(resp, sz, sub->ppath, safestrlen(sub->ppath));
						EddIntData(resp, sz, ev->etype);
						EddHexData(resp, sz, (const char *)&(ev->printer_up_time), sizeof(time_t));
						EddIntData(resp, sz, sub->seqno); /*rfc3996 says seqno from subscription, not event  EddIntData(resp, sz, ev->seqno);*/
						EddHexData(resp, sz, sub->user_data, sub->ulen);
						EddIntData(resp, sz, ev->aux_id);
						EddIntData(resp, sz, ev->state);
						EddHexData(resp, sz, ev->reason, safestrlen(ev->reason));
						EddHexData(resp, sz, ev->xname, safestrlen(ev->xname));
						/*status ? not needed yet*/

						/*DEBUG2("Ips_get_notifications: list sub id %d event seqno %d printer %s src %s state %d id %d xname %s",
						        sub->id, ev->seqno, ev->printer ? ev->printer : "<NULL>", ev->etype & (IEVENT_PR_STATECHANGE | IEVENT_PR_STOPPED) ? "printer" : "job",
						         ev->state, ev->aux_id, ev->xname ? ev->xname : "<NULL>" );*/
					}
				}
				if (!waitinterval) {
					for (cl = sub->clients; cl; cl = cl->next) {
						if (cl->i == fd) goto clex;
					}
					cl = malloc_or_die(sizeof (struct int_list), __FILE__, __LINE__);
					cl->i = fd;
					cl->next = sub->clients;
					sub->clients = cl;
				  clex: ;
				}
				goto subd_found;
			}
		}
	  subd_found:
		;
	}

    endproc:

	/*
	if (printer) free(printer);
	if (ppath) free(ppath);
	*/

	EddLf(resp, sz);
	DEBUG2("Ips_get_notifications: response: %s", resp);
	DEBUG1("Ips_get_notifications: operation status %d, event count %d", status, evcount);
	/*send response back*/
	status = write(fd, resp, safestrlen(resp));
	DEBUG2("Ips_get_notifications: send response fd %d write bytes %d errno %d", fd, status >=0 ? status : 0, status >= 0 ? 0 : errno);

	if (resp) free(resp);

	return 0;
}


int Ipe_printer_state_changed(const char *printer, int state, int job_accept) {

	char buf[SMALLBUFFER];
	int status, e;

	buf[0] = '\0';
	EddIntData(buf, sizeof(buf), E_PRS_CHANGE);
	EddHexData(buf, sizeof(buf), printer, safestrlen(printer));
	EddIntData(buf, sizeof(buf), state);
	if (state == IPPAV_PRS_STOPPED)
		EddHexData(buf, sizeof(buf), IPPAV_PRSR_PAUSED, safestrlen(IPPAV_PRSR_PAUSED));
	else
		EddHexData(buf, sizeof(buf), IPPAV_PRSR_NONE, safestrlen(IPPAV_PRSR_NONE));
	EddIntData(buf, sizeof(buf), job_accept);
	EddHexData(buf, sizeof(buf), "", 0); /*xname*/
	EddLf(buf, sizeof(buf));

	DEBUG2("Ipe_printer_state_changed: printer %s, state %d, reason %s, job_accept %d, send string buf %s", 
	        printer, state, state == IPPAV_PRS_STOPPED ? IPPAV_PRSR_PAUSED : IPPAV_PRSR_NONE, job_accept, buf);
	/*send to server process*/
	status = write(Evsrv_fd, buf, safestrlen(buf));
	if (status < safestrlen(buf)) {
		e = errno;
		DEBUG2("Ipe_printer_state_changed: Evsrv_fd write error, sent %d bytes, errno %d", status, status == 0 ? e : 0);
		return -1;
	}

	return 0;
}

int Ipe_job_changed(const char *printer, int event, int state, struct job *job) {

	char buf[SMALLBUFFER], *s;
	int status, e;

	buf[0] = '\0';
	EddIntData(buf, sizeof(buf), event == E_JOB_CREATE ? E_JOB_CREATE : E_JOB_CHANGE);
	EddHexData(buf, sizeof(buf), printer, safestrlen(printer));
	EddIntData(buf, sizeof(buf), state);
	EddHexData(buf, sizeof(buf), IPPAV_JSR_NONE, safestrlen(IPPAV_JSR_NONE));
	EddIntData(buf, sizeof(buf), Find_decimal_value(&job->info, NUMBER));
	s = Find_str_value(&job->info, JOBNAME);
	EddHexData(buf, sizeof(buf), s, safestrlen(s)); /*xname*/
	EddLf(buf, sizeof(buf));

	DEBUG2("Ipe_job_changed: printer %s, state %d reason %s job_id %d job_name %s send string buf %s",
	        printer, state, IPPAV_JSR_NONE, Find_decimal_value(&job->info, NUMBER), Find_str_value(&job->info, JOBNAME), buf);
	/*send to server process*/
	status = write(Evsrv_fd, buf, safestrlen(buf));
	if (status < safestrlen(buf)) {
		e = errno;
		DEBUG2("Ipe_job_changed: Evsrv_fd write error, sent %d bytes, errno %d", status, status == 0 ? e : 0);
		return -1;
	}

	return 0;
}



int Ipe_lpc(int ewid, int erpipe, struct line_list *params) 
{

	char buf[SMALLBUFFER], *c, *d;
	int lpc;
	int trans;		/*transaction id*/
	int bytes;
	size_t sz;
	int status, rs;
	fd_set readfd;
	struct timeval timeval;
	struct line_list t;
	int e;
	int rcount;

	trans = time(0); 
	memset(buf, 0, sizeof(buf));
	rcount = 0;

	DEBUGFC(DNW4)Dump_line_list("Ipe_lpc: params", params);

	lpc = Find_decimal_value(params, KWC_PROC);
	EddIntData(buf, sizeof(buf), lpc);
	EddIntData(buf, sizeof(buf), ewid);
	EddIntData(buf, sizeof(buf), trans);
	c = Find_str_value(params, KWA_PRINTER); EddHexData(buf, sizeof(buf), c, safestrlen(c));
	c = Find_str_value(params, KWA_PPATH); EddHexData(buf, sizeof(buf), c, safestrlen(c));
	c = Find_str_value(params, KWA_USER); EddHexData(buf, sizeof(buf), c, safestrlen(c));
	switch (lpc) {
		case E_SUBS_CREATE:
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_LEASE));
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_EVENTS));
			c = Find_str_value(params, KWC_USERDATA); 
			ExmDataHex(&d, c); EddHexData(buf, sizeof(buf), d, safestrlen(d)); if (d) free(d);
			break;
		case E_SUBS_ATTRS:
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_SUBS_ID));
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_JOBID));
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_LIMIT));
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_MYSUBS));
			break;
		case E_SUBS_RENEW:
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_SUBS_ID));
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_LEASE));
			break;
		case E_SUBS_CANCEL:
			EddIntData(buf, sizeof(buf), Find_decimal_value(params, KWC_SUBS_ID));
			break;
		default:
			DEBUGF(DNW2)("Ipe_lpc: unknown proc call %d", lpc);
			return SERVER_ERROR_INTERNAL_ERROR;
	}
	EddLf(buf, sizeof(buf));

	DEBUGF(DNW4)("Ipe_lpc: lpc send string buf %s", buf);
	/*send to server process*/
	status = write(Evsrv_fd, buf, safestrlen(buf));
	if (status < safestrlen(buf)) {
		e = errno;
		DEBUGF(DNW2)("Ipe_lpc: Evsrv_fd write error, sent %d bytes, errno %d", status, status == 0 ? e : 0);
		return SERVER_ERROR_INTERNAL_ERROR;
	}

  readmsg:
 
	/*read response - size*/
	memset(&timeval, 0, sizeof(timeval));
	timeval.tv_sec = NTF_SP_TIMEOUT;
	FD_ZERO(&readfd);
	FD_SET(erpipe, &readfd);
	if ( (status = select(erpipe+1, &readfd, NULL, NULL, &timeval)) <= 0) {
		if (!status) {
			DEBUGF(DNW2)("Ipe_lpc: server read response size timeout");
			return SERVER_ERROR_BUSY;
		} else {
			e = errno;
			DEBUGF(DNW2)("Ipe_lpc: server read response size error %d", e);
			return SERVER_ERROR_INTERNAL_ERROR;
		}
	}
	bytes = ok_read(erpipe, buf, 2*sizeof(size_t));
	if (bytes >=0) buf[bytes] = '\0'; else buf[0] = '\0';
	sz = strtoul(buf, NULL, 16); /*hope long is sufficient*/
	if (sz <= 0) sz = SMALLBUFFER;

	/*response data*/
	memset(&timeval, 0, sizeof(timeval));
	timeval.tv_sec = NTF_SP_TIMEOUT;
	FD_ZERO(&readfd);
	FD_SET(erpipe, &readfd);
	if ( (status = select(erpipe+1, &readfd, NULL, NULL, &timeval)) <= 0) {
		if (!status) {
			DEBUGF(DNW2)("Ipe_lpc: server read response data timeout");
			return SERVER_ERROR_BUSY;
		} else {
			e = errno;
			DEBUGF(DNW2)("Ipe_lpc: server read response data error %d", e);
			return SERVER_ERROR_INTERNAL_ERROR;
		}
	}

	d = malloc_or_die(sz + 1, __FILE__, __LINE__);
	bytes = ok_read(erpipe, d, sz); d[sz] = '\0';

	rs = SERVER_ERROR_INTERNAL_ERROR;
	Init_line_list(&t);
	Split(&t,d,Whitespace,0,0,0,' ',0,0);
	if ((c = strrchr(d, '\n'))) *c = '\0';
	DEBUGF(DNW4)("Ipe_lpc: server response %s len %lu/%d", d, (unsigned long) sz, safestrlen(d));
	DEBUGFC(DNW4)Dump_line_list("Ipe_lpc: response linelist", &t);
	if ((t.count) >=1) {
		switch (strtol(t.list[0], 0, 0)) {
			case E_NONE:/*read next response*/
				if (d) free(d);
				Free_line_list(&t);
				if (++rcount < 16) goto readmsg;
				break;
			case E_SUBS_CREATE:
				if (t.count < 5) break;
				Set_decimal_value(params, KWR_STATUS, strtol(t.list[2], 0, 0));
				Set_decimal_value(params, KWR_SUBS_ID, strtol(t.list[3], 0, 0));
				Set_decimal_value(params, KWR_LEASE, strtol(t.list[4], 0, 0));
				rs = SUCCESSFUL_OK;
				break;
			case E_SUBS_ATTRS:
				if (t.count < 3) break;
				rs = strtol(t.list[2], 0, 0);
				Free_line_list(params);
				for (int i = 3; i < t.count; i++) Add_line_list(params, t.list[i] , NULL, 0, 0);
				break;
			case E_SUBS_RENEW:
				if (t.count < 3) break;
				Set_decimal_value(params, KWR_STATUS, strtol(t.list[2], 0, 0));
				Set_decimal_value(params, KWR_LEASE, strtol(t.list[3], 0, 0));
				rs = SUCCESSFUL_OK;
				break;
			case E_SUBS_CANCEL:
				if (t.count < 2) break;
				Set_decimal_value(params, KWR_STATUS, strtol(t.list[2], 0, 0));
				rs = SUCCESSFUL_OK;
				break;
			default:
				break;
		}
	}

	if (d) free(d);
	Free_line_list(&t);

	return rs;
}
