/***************************************************************************
 * LPRng - An Extended Print Spooler System
 *
 * IPP internet printing protocol 
 * copyright 2021 Vaclav Michalek
 * 
 * See LICENSE for conditions of use.
 ***************************************************************************/


#ifndef _IPP_EVENTS_H_
#define _IPP_EVENTS_H_ 


/* CONST MACROS*/

#define NTF_LEASE_DURATION_MIN	30
#define NTF_LEASE_DURATION_MAX	(Done_jobs_max_age_DYN > 7200 ? Done_jobs_max_age_DYN : 7200 )
#define NTF_LEASE_DURATION	(Done_jobs_max_age_DYN > NTF_LEASE_DURATION_MIN ? Done_jobs_max_age_DYN : 1800)  /*CUPS:2147483647,default 86400, gsd-print-notifications requires 3600 */
#define NTF_EVENT_LIFE		120   /*(Done_jobs_max_age_DYN > 20 ? Done_jobs_max_age_DYN >> 1 : 10)*/  /*CUPS:15*/
#define NTF_MAX_SUBSCRIPTIONS	128                                                                       /*CUPS:100*/
#define NTF_MAX_EVNT_COUNT	512	/*limit number of events in subscription object*/
#define NTF_MAX_EVNT_TYPES	6	/*max number of event type in subscription create */
#define NTF_MIN_WAIT		10	/*minimum notify-get-interval*/                                   /*CUPS:10-30 by load and state*/
#define NTF_SP_TIMEOUT		10/*60*//*1*/	/*subscription lpc timeout seconds*/

/*keys for params of subscription lpc*/
EXTERN const char * KWC_PROC		DEFINE ( = "LPC");
EXTERN const char * KWC_SUBS_ID		DEFINE ( = "LSUBSID");
EXTERN const char * KWC_LEASE		DEFINE ( = "LEASE");
EXTERN const char * KWC_EVENTS		DEFINE ( = "LEVENTS");
EXTERN const char * KWC_USERDATA	DEFINE ( = "LUSERDATA");
EXTERN const char * KWC_LIMIT		DEFINE ( = "LLIMIT");
EXTERN const char * KWC_MYSUBS		DEFINE ( = "LMYSUBS");
EXTERN const char * KWC_JOBID		DEFINE ( = "LJOBID");
EXTERN const char * KWR_STATUS		DEFINE ( = "RSTATUS");
EXTERN const char * KWR_SUBS_ID		DEFINE ( = "RSUBSID");
EXTERN const char * KWR_LEASE		DEFINE ( = "RLEASE");

/*ipp internal event bitmask*/
#define IEVENT_PRINTER		0x0000
#define IEVENT_PR_STATECHANGE	0x0001
#define IEVENT_PR_STOPPED	0x0002
#define IEVENT_JOB_STATECHANGE	0x0010
#define IEVENT_JOB_CREATED	0x0020
#define IEVENT_JOB_COMPLETED	0x0040


/* DATATYPES */

/*subscriptions*/

struct ipp_event {
	int seqno;
	int etype;
	char *printer;
	time_t printer_up_time;
	int aux_id;		/*job id or printer is accepting jobs; we are lazy to use union*/
	int state;		/*job or printer state*/
	char *reason;
	char *xname;		/*job name*/
	struct ipp_event *next;
};

struct int_list {
	int i;
	struct int_list *next;
};

struct ipp_subs {
	int id;			/*subscription id*/
	struct int_list *clients;	/*fds of running notification clients*/
	char *printer;
	char *ppath;
	char *username;
	int seqno;		/*notify seqence number*/
	time_t expire;		/*expiration time*/
	int lease;
	int event_types;	/*subscribed events*/
	char *user_data;		/*notify-user-data text of hex*/
	int ulen;
	struct ipp_event *events;
	struct ipp_subs *next;
};

/* GLOBAL VARIABLES*/

EXTERN struct ipp_subs *ipp_subs_list DEFINE (= NULL);


/* PROTOTYPES */


int Ips_remove_subscription(const char *printer, const char *ppath, int sub_id);
int Ips_delete_wait_client(int fd);

int Ips_subscription_create(struct line_list *params);
int Ips_subscription_attributes(struct line_list *params);
int Ips_subscription_renew(struct line_list *params);
int Ips_subscription_cancel(struct line_list *params);
int Ips_event(struct line_list *params);
int Ips_get_notifications(struct line_list *params);

int Ipe_printer_state_changed(const char *printer, int state, int job_accept);
int Ipe_job_changed(const char *printer, int event, int state, struct job *job);
int Ipe_lpc(int ewid, int erpipe, struct line_list *params);


#endif
